from lunespy.utils.crypto.converters import bytes_to_string
from lunespy.utils.crypto.converters import string_to_bytes
from lunespy.client.transactions.constants import BurnType
from lunespy.utils.crypto.converters import sign
from lunespy.utils import bcolors
from lunespy.client.wallet import Account
from lunespy.utils import now
from base58 import b58decode
from requests import post
import struct


def mount_burn(sender: Account, burn_data: dict) -> dict:
    timestamp: int = burn_data.get('timestamp', int(now() * 1000))
    burn_fee: int = burn_data.get('burn_fee', BurnType.fee.value)
    quantity: int = burn_data.get('quantity', 0)
    asset_id: int = burn_data['asset_id']

    bytes_data: bytes = BurnType.to_byte.value + \
        b58decode(sender.public_key) + \
        b58decode(asset_id) + \
        struct.pack(">Q", quantity) + \
        struct.pack(">Q", burn_fee) + \
        struct.pack(">Q", timestamp)

    signature: bytes = sign(sender.private_key, bytes_data)
    mount_tx = {
        "type": BurnType.to_int.value,
        "senderPublicKey": sender.public_key,
        "signature": signature.decode(),
        "timestamp": timestamp,
        "fee": burn_fee,

        "assetId": asset_id,
        "quantity": quantity
    }
    return mount_tx


def validate_burn(sender: Account, burn_data: dict) -> bool:
    quantity: int = burn_data.get('quantity', -1)
    asset_id: str = burn_data.get('asset_id', False)

    if not sender.private_key:
        print(bcolors.FAIL + 'Burner `Account` not have a private key' + bcolors.ENDC)
        return False

    if quantity < 0:
        print(bcolors.FAIL + 'Burn `quantity` cannot be less than 0' + bcolors.ENDC)
        return False

    if asset_id == False:
        print(bcolors.FAIL + 'Burn_data `asset_id` cannot exists' + bcolors.ENDC)
        return False

    return True


# todo async
def send_burn(mount_tx: dict, node_url: str) -> dict:
    response = post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
        })

    if response.ok:
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx
    else:
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx



from lunespy.client.transactions.constants import LeaseType
from lunespy.utils.crypto.converters import sign
from lunespy.utils import bcolors
from lunespy.client.wallet import Account

from lunespy.utils import now
from base58 import b58decode
from requests import post
import struct


def mount_lease(sender: Account, validator_address: str, lease_data: dict) -> dict:
    timestamp: int = lease_data.get('timestamp', int(now() * 1000))
    amount: int = lease_data['amount']
    lease_fee: int = lease_data.get('lease_fee', LeaseType.fee.value)

    bytes_data: bytes = LeaseType.to_byte.value + \
        b58decode(sender.public_key) + \
        b58decode(validator_address) + \
        struct.pack(">Q", amount) + \
        struct.pack(">Q", lease_fee) + \
        struct.pack(">Q", timestamp)

    signature: bytes = sign(sender.private_key, bytes_data)
    mount_tx: dict = {
        "type": LeaseType.to_int.value,
        "senderPublicKey": sender.public_key,
        "signature": signature.decode(),
        "timestamp": timestamp,
        "fee": lease_fee,

        "recipient": validator_address,
        "amount": amount
    }
    return mount_tx


def validate_lease(sender: Account, lease_data: dict) -> bool:
    amount: int = lease_data.get('amount', -1)

    if not sender.private_key:
        print(bcolors.FAIL + 'Staker `Account` not have a private key' + bcolors.ENDC)
        return False
    elif amount <= 0:
        print(bcolors.FAIL + 'Leasing `amount` cannot be less than 0' + bcolors.ENDC)
        return False
    return True


# todo async
def send_lease(mount_tx: dict, node_url: str) -> dict:
    response = post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
        })

    if response.ok:
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx
    else:
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx



from lunespy.client.transactions import BaseTransaction
from lunespy.client.transactions.alias.validators import validate_alias
from lunespy.client.transactions.alias.validators import mount_alias
from lunespy.client.transactions.alias.validators import send_alias
from lunespy.utils import bcolors
from lunespy.client.wallet import Account

class CreateAlias(BaseTransaction):
    """
    alias_data: dict
        timestamp: int    
        alias_fee: int
        alias: str
    """
    def __init__(self, sender: Account, **alias_data: dict) -> None:
        super().__init__('Alias', alias_data)
        self.sender: Account = sender
        self.alias_data: dict = alias_data
        self.history: list = []

   
    @property
    def ready(self) -> bool:
        return validate_alias(self.sender, self.alias_data)
    
    @property
    def transaction(self) -> dict:
        return super().transaction(
            mount_tx=mount_alias,
            sender=self.sender,
            alias_data=self.alias_data)

    def send(self, node_url: str = None) -> dict:
        tx = super().send(send_alias, node_url)
        self.history.append(tx)
        return tx


from lunespy.client.transactions.burn.validators import validate_burn
from lunespy.client.transactions.burn.validators import mount_burn
from lunespy.client.transactions.burn.validators import send_burn
from lunespy.client.transactions import BaseTransaction
from lunespy.client.wallet import Account



class BurnToken(BaseTransaction):
    """
    burn_data: dict
        timestamp: int 
        asset_id: str 
        quantity: int
        burn_fee: int
    """
    def __init__(self, sender: Account, **burn_data: dict) -> None:
        super().__init__('Burn Token', burn_data)
        self.sender = sender
        self.burn_data = burn_data
        self.burn_data['token_type'] = 'Token'
        self.history = []

    @property
    def ready(self) -> bool:
        return validate_burn(self.sender, self.burn_data)
    
    @property
    def transaction(self) -> dict:
        return super().transaction(
            mount_tx=mount_burn,
            sender=self.sender,
            burn_data=self.burn_data)

    def send(self, node_url: str = None) -> dict:
        tx = super().send(send_burn, node_url)
        self.history.append(tx)
        return tx


class BurnAsset(BurnToken):
    def __init__(self, sender: Account, **burn_data: dict) -> None:
        burn_data['token_type'] = burn_data['token_type'] if burn_data.get('token_type', False) else 'Asset'
        BaseTransaction.__init__(self, tx_type='Burn Asset', tx_data=burn_data)
        BurnToken.__init__(self, sender=sender, **burn_data)


class BurnNFT(BurnAsset):
    def __init__(self,
        sender: Account,
        **burn_data: dict
        ) -> None:
        burn_data['token_type'] = 'NFT'
        BaseTransaction.__init__(self, tx_type='Burn NFT', tx_data=burn_data)
        BurnToken.__init__(self, sender=sender, **burn_data)



from lunespy.client.transactions.cancel_lease.validators import validate_cancel
from lunespy.client.transactions.cancel_lease.validators import mount_cancel
from lunespy.client.transactions.cancel_lease.validators import send_cancel
from lunespy.client.transactions import BaseTransaction
from lunespy.client.wallet import Account


class CancelLease(BaseTransaction):
    """
    cancel_data: dict
        lease_tx_id: str
        cancel_fee: int
        timestamp: int
    """
    def __init__(self, sender: Account, **cancel_data: dict) -> None:
        super().__init__('Create Lease', cancel_data)
        self.sender: Account = sender
        self.cancel_data: dict = cancel_data
        self.history: list = []

    
    @property
    def ready(self) -> bool:
        return validate_cancel(self.sender, self.cancel_data)


    @property
    def transaction(self) -> dict:
        return super().transaction(
            mount_tx=mount_cancel,
            sender=self.sender,
            cancel_data=self.cancel_data)

    def send(self, node_url: str = None) -> dict:
        tx = super().send(send_cancel, node_url)
        self.history.append(tx)
        return tx


from lunespy.client.transactions.constants import ReissueType
from lunespy.utils.crypto.converters import sign
from lunespy.utils import bcolors
from lunespy.client.wallet import Account
from lunespy.utils import now
from base58 import b58decode
from requests import post
import struct

def validate_reissue(sender: Account, reissue_data: dict) -> bool:
    quantity: int = reissue_data.get('quantity', -1)
    asset_id: str = reissue_data.get('asset_id', False)

    if not sender.private_key:
        print(bcolors.FAIL + 'Creator `Account` not have a private key' + bcolors.ENDC)
        return False

    if quantity < 0:
        print(bcolors.FAIL + 'To Reissue the `quantity` cannot be less than 0' + bcolors.ENDC)
        return False

    if asset_id == False:
        print(bcolors.FAIL + 'To Reissue pass an `asset_id`' + bcolors.ENDC)
        return False

    return True


def mount_reissue(sender: Account, reissue_data: dict) -> dict:
    reissue_fee = reissue_data.get('reissue_fee', ReissueType.fee.value)
    timestamp = reissue_data.get('timestamp', int(now() * 1000))
    reissuable = reissue_data.get('reissuable', False)
    asset_id = reissue_data['asset_id']
    quantity = reissue_data['quantity']

    bytes_data = ReissueType.to_byte.value + \
        b58decode(sender.public_key) + \
        b58decode(asset_id) + \
        struct.pack(">Q", quantity) + \
        (b'\1' if reissuable else b'\0') + \
        struct.pack(">Q",reissue_fee) + \
        struct.pack(">Q", timestamp)

    signature: bytes = sign(sender.private_key, bytes_data)
    mount_tx = {
        "type": ReissueType.to_int.value,
        "senderPublicKey": sender.public_key,
        "signature": signature.decode(),
        "timestamp": timestamp,
        "fee": reissue_fee,

        "assetId": asset_id,
        "reissuable": reissuable,
        "quantity": quantity
    }
    return mount_tx


# todo async
def send_reissue(mount_tx: dict, node_url: str) -> dict:
    response = post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
        })

    if response.ok:
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx
    else:
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx



from lunespy.client.transactions.lease.validators import validate_lease
from lunespy.client.transactions.lease.validators import mount_lease
from lunespy.client.transactions.lease.validators import send_lease
from lunespy.client.transactions import BaseTransaction
from lunespy.client.wallet import Account


class CreateLease(BaseTransaction):
    """
    lease_data: dict
        validator_address: str
        lease_fee: int
        timestamp: int
        amount: int
    """
    def __init__(self, sender: Account, validator_address: str, **lease_data: dict) -> None:
        super().__init__('Create Lease', lease_data)
        self.sender: Account = sender
        self.validator_address: str = validator_address
        self.lease_data: dict = lease_data
        self.history: list = []

    
    @property
    def ready(self) -> bool:
        return validate_lease(self.sender, self.lease_data)


    @property
    def transaction(self) -> dict:
        return super().transaction(
            mount_tx=mount_lease,
            sender=self.sender,
            validator_address=self.validator_address,
            lease_data=self.lease_data)

    def send(self, node_url: str = None) -> dict:
        tx = super().send(send_lease, node_url)
        self.history.append(tx)
        return tx


from lunespy.client.transactions import BaseTransaction
from lunespy.client.transactions.reissue.validators import validate_reissue
from lunespy.client.transactions.reissue.validators import mount_reissue
from lunespy.client.transactions.reissue.validators import send_reissue
from lunespy.utils import bcolors
from lunespy.client.wallet import Account


class ReissueToken(BaseTransaction):
    """
    reissue_data: dict
        reissuable: bool
        reissue_fee: int
        asset_id: str
        quantity: int
    """
    def __init__(self, sender: Account, **reissue_data: dict) -> None:
        super().__init__('Reissue Token', reissue_data)
        self.sender = sender
        self.reissue_data = reissue_data
        self.reissue_data['token_type'] = 'Token'
        self.history = []

    @property
    def ready(self) -> bool:
        return validate_reissue(self.sender, self.reissue_data)
    
    @property
    def transaction(self) -> dict:
        return super().transaction(
            mount_tx=mount_reissue,
            sender=self.sender,
            reissue_data=self.reissue_data)

    def send(self, node_url: str = None) -> dict:
        tx = super().send(send_reissue, node_url)
        self.history.append(tx)
        return tx


class ReissueAsset(ReissueToken):
    def __init__(self, sender: Account, **reissue_data: dict) -> None:
        reissue_data['token_type'] = 'Asset'
        BaseTransaction.__init__(self, tx_type='Reissue Asset', tx_data=reissue_data)
        ReissueToken.__init__(self, sender=sender, **reissue_data)


class ReissueNFT(ReissueToken):
    def __init__(self, sender: Account, **reissue_data: dict) -> None:
        reissue_data['token_type'] = 'NFT'
        reissue_data['decimals'] = 0
        BaseTransaction.__init__(self, tx_type='Reissue NFT', tx_data=reissue_data)
        ReissueToken.__init__(self, sender=sender, **reissue_data)



from lunespy.client.transactions.transfer.validators import validate_transfer
from lunespy.client.transactions.transfer.validators import mount_transfer
from lunespy.client.transactions.transfer.validators import send_transfer
from lunespy.client.transactions import BaseTransaction
from lunespy.client.wallet import Account

class TransferToken(BaseTransaction):
    """
    transfer_data: dict
        transfer_fee: int
        timestamp: int
        asset_fee: int
        asset_id: str
        amount: int
    """
    def __init__(self, sender: Account, receiver: Account, **transfer_data: dict) -> None:
        super().__init__('Transfer', transfer_data)
        self.sender: Account = sender
        self.receiver: Account = receiver
        self.transfer_data: dict = transfer_data
        self.history: list = []

    @property
    def ready(self) -> bool:
        return validate_transfer(self.sender, self.receiver, self.transfer_data)

    @property
    def transaction(self) -> dict:
        return super().transaction(
            mount_tx=mount_transfer,
            sender=self.sender,
            receiver=self.receiver,
            transfer_data=self.transfer_data)

    def send(self, node_url: str = None) -> dict:
        tx = super().send(send_transfer, node_url)
        self.history.append(tx)
        return tx



from lunespy.utils import export_json
from lunespy.utils import log_data
from lunespy.utils import bcolors
from lunespy.server import Node
from abc import abstractmethod
from abc import ABCMeta


class BaseTransaction(metaclass=ABCMeta):
    def __init__(self, tx_type: str, tx_data: dict):
        self.tx_type: str = tx_type
        self.tx_data: str = tx_data

    @abstractmethod
    def ready(self) -> bool:
        raise NotImplementedError


    def transaction(self, mount_tx, **extra) -> dict:
        if self.ready:
            tx = {'ready': True}
            tx.update( mount_tx( **extra ) )
            return tx
        else:
            print(bcolors.FAIL + f'{self.tx_type} Transactions bad formed', bcolors.ENDC)
            return {'ready': False}


    def send(self, send_tx, node_url: str) -> dict:
        if node_url == None:
            if self.sender.network == 'mainnet':
                node_url = Node.mainnet_url.value
            else:
                node_url = Node.testnet_url.value

        mounted_tx = self.transaction
        if mounted_tx['ready']:
            tx_response = send_tx(mounted_tx, node_url=node_url)

            if tx_response['send']:
                self.show(**tx_response)
                return tx_response
            else:
                print(bcolors.FAIL + f"Your {self.tx_type} dont sended because:\n└──{tx_response['response']}" + bcolors.ENDC)
                return tx_response

        else:
            print(bcolors.FAIL + f'{self.tx_type} Transaction dont send', bcolors.ENDC)
            return mounted_tx
        

    def show(self, **data: dict) -> None:
        log_data(data)
        export_json(data)



from lunespy.client.transactions.constants import CancelLeaseType
from lunespy.utils.crypto.converters import sign
from lunespy.utils import bcolors
from lunespy.client.wallet import Account

from lunespy.utils import now
from base58 import b58decode
from requests import post
import struct


def mount_cancel(sender: Account, cancel_data: dict) -> dict:
    lease_tx_id: str = cancel_data['lease_tx_id']
    timestamp: int = cancel_data.get('timestamp', int(now() * 1000))
    cancel_fee: int = cancel_data.get('cancel_fee', CancelLeaseType.fee.value)

    bytes_data: bytes = CancelLeaseType.to_byte.value + \
        b58decode(sender.public_key) + \
        struct.pack(">Q", cancel_fee) + \
        struct.pack(">Q", timestamp) + \
        b58decode(lease_tx_id)

    signature: bytes = sign(sender.private_key, bytes_data)
    mount_tx: dict = {
        "type":CancelLeaseType.to_int.value,
        "senderPublicKey": sender.public_key,
        "signature": signature.decode(),
        "timestamp": timestamp,
        "fee": cancel_fee,

        "leaseId": lease_tx_id
    }
    return mount_tx


def validate_cancel(sender: Account, cancel_data: dict) -> bool:
    amount: int = cancel_data.get('amount', -1)
    lease_tx_id: str = cancel_data.get('lease_tx_id', '')

    if not sender.private_key:
        print(bcolors.FAIL + 'Staker `Account` not have a private key' + bcolors.ENDC)
        return False
    elif lease_tx_id == '':
        print(bcolors.FAIL + 'Leasing must be `lease_tx_id`' + bcolors.ENDC)
        return False
    return True


# todo async
def send_cancel(mount_tx: dict, node_url: str) -> dict:
    response = post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
        })

    if response.ok:
        mount_tx['send'] = True
        mount_tx['response'] = response.json()
        return mount_tx
    else:
        mount_tx['send'] = False
        mount_tx['response'] = response.text
        return mount_tx


from lunespy.client.transactions.issue.validators import validate_issue
from lunespy.client.transactions.issue.validators import send_issue
from lunespy.client.transactions.issue.validators import mount_issue
from lunespy.client.transactions import BaseTransaction
from lunespy.utils import bcolors
from lunespy.client.wallet import Account


class IssueToken(BaseTransaction):
    """
    issue_data: dict
        description: str
        reissuable: bool
        issue_fee: int
        quantity: int
        decimals: int
        name: str
    """
    def __init__(self, sender: Account, **issue_data: dict) -> None:
        super().__init__('Issue Token', issue_data)
        self.sender = sender
        self.issue_data = issue_data
        self.issue_data['token_type'] = 'Token'
        self.history = []


    @property
    def ready(self) -> bool:
        return validate_issue(self.sender, self.issue_data)
    
    @property
    def transaction(self) -> dict:
        return super().transaction(
            mount_tx=mount_issue,
            sender=self.sender,
            issue_data=self.issue_data)

    def send(self, node_url: str = None) -> dict:
        tx = super().send(send_issue, node_url)
        self.history.append(tx)
        return tx

class IssueAsset(IssueToken):
    def __init__(self, sender: Account, **issue_data: dict) -> None:   
        issue_data['token_type'] = 'Asset'
        IssueToken.__init__(self, sender=sender, **issue_data)
        BaseTransaction.__init__(self, tx_type='Issue Asset', tx_data=issue_data)
    

class IssueNFT(IssueToken):
    def __init__(self, sender: Account, **issue_data: dict) -> None:
        issue_data['token_type'] = 'NFT'
        issue_data['decimals'] = 0
        IssueToken.__init__(self, sender=sender, **issue_data)
        BaseTransaction.__init__(self, tx_type='Issue NFT', tx_data=issue_data)



class Account:
    """
    params: 
        private_key: str
        public_key: str
        hash_seed: str
        network_id: str | '0' xor '1'
        address: str
        nonce: int | 0 to 4.294.967.295
        network: str | 'mainnet' xor 'testnet'
        seed: str | 12 words
        byte_private_key: bytes 
        byte_public_key: bytes
        byte_address: bytes
    """
    def __init__(self, **wallet: dict):
        from lunespy.client.wallet.validators import validate_wallet
        
        self.__dict__ = validate_wallet(wallet)


    def __str__(self) -> str:
        from lunespy.utils import bcolors
        data = ''
        for key, value in self.__dict__.items():
            if 'byte' not in key and 'id' not in key:
                data += f"\n{key}{bcolors.OKGREEN}\n └── {value}{bcolors.ENDC}"
        return data


    def __repr__(self) -> str:
        data = ''
        for key, value in self.__dict__.items():
                data += f"{key} -> {value}\n"
        return data


    def to_json(self, path: str = '.') -> str:
        from lunespy.utils import export_json
        data = {
            key: value 
            for (key, value) in self.__dict__.items()
            if 'byte' not in key and 'id' not in key
        }

        return export_json(
            data=data,
            name=f'wallet-{self.network}',
            path=path
        )



from requests import get

def block_from_height(node_url: str, height: int) -> dict:
    full_url = f'https://{node_url}/blocks/at/{height}'
    response = get(full_url)
    
    if response.ok:
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def range_block(node_url: str, start_block: int, end_block: int) -> dict:
    full_url = f'https://{node_url}/blocks/seq/{start_block}/{end_block}'
    response = get(full_url)

    if response.ok:
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def last_block(node_url: str) -> dict:
    full_url = f'https://{node_url}/blocks/last'
    response = get(full_url)

    if response.ok:
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
                'status': 'error',
                'response': response.text
            }


def blocks_generated_by_specified_address(node_url: str, address: str, start_block: int, end_block: int) -> dict:
    full_url = f'https://{node_url}/blocks/address/{address}/{start_block}/{end_block}'
    response = get(full_url)

    if response.ok:
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }
